/*-
 * Copyright 2013 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_video.h"
#include "iicbus_if.h"
#include "cx88_i2c_if.h"
#include "cx23880_video_regs.h"

typedef struct
{
   uint32_t sample_freq;
   uint32_t hactive;
   uint32_t hdelay;
   uint32_t vactive;
   uint32_t vdelay;
   uint32_t vbi_packets;
   uint16_t pixels_per_line;
} cx23880_video_spec;

typedef struct
{
   cx88_common_softc common;
   cx88_i2c_softc i2c;
   cx88_video_softc video;
   const cx23880_video_spec *spec;
} cx23880_video_softc;

static cx88_i2c_softc *cx23880_video_i2c(cx88_common_softc *sc, uint8_t bus);
static int cx23880_video_setup(cx88_common_softc *arg);
static int cx23880_video_teardown(cx88_common_softc *arg);
static int cx23880_video_filter(void *arg);
static void cx23880_video_intr(void *arg);
static int cx23880_video_probe(device_t dev);
static int cx23880_video_attach(device_t dev);
static int cx23880_video_shutdown(device_t dev);
static int cx23880_video_suspend(device_t dev);
static cx88_i2c_bus *cx23880_video_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx23880_video_get_i2c_dev(device_t dev, uint8_t addr);
static int cx23880_video_request_reconfig(device_t dev); 

#define CX23880_VIDEO_SPECS CX88_VIDEO_FMT_SECAM
#define CX23880_FMT_TO_SPEC(fmt) (fmt - 1)

static const cx23880_video_spec specs[CX23880_VIDEO_SPECS] = 
{
   /*sample_freq  hactive  hdelay   vactive  vdelay   vbi_packets    pixels_per_line*/
   {28636363,     754,     135,     480,     24,      400,           910},  /* NTSC-M */
   {28636363,     754,     135,     480,     24,      400,           910},  /* NTSC-J */
   {35468950,     922,     186,     480,     24,      511,           1126}, /* NTSC-4.43 */
   {35468950,     922,     186,     576,     36,      511,           1135}, /* PAL-B/D/G/H/I */
   {28604888,     754,     135,     480,     24,      400,           909},  /* PAL-M */
   {28656448,     754,     135,     576,     36,      400,           917},  /* PAL-N */
   {28656448,     754,     135,     576,     36,      400,           917},  /* PAL-Nc */
   {35468950,     922,     186,     480,     24,      511,           1126}, /* PAL-60 */
   {35468950,     922,     186,     576,     36,      511,           1135}  /* SECAM */
};

static const cx88_client_driver cx23880_video_client = 
{
   .softc_size               = sizeof(cx23880_video_softc),
   .bridge                   = CX88_BRIDGE_CX23880,
   .function                 = CX88_FUNC_VIDEO,
   .setup                    = cx23880_video_setup,
   .teardown                 = cx23880_video_teardown,
	.interrupt_filter         = cx23880_video_filter,
   .interrupt_handler        = cx23880_video_intr,
   .description              = "Conexant CX23880/1/2/3 Analog Video"
};

static cx88_i2c_softc *cx23880_video_i2c(cx88_common_softc *sc, uint8_t bus)
{
   if (bus != 0)
   {
      return NULL;
   }
   return &(((cx23880_video_softc*)sc)->i2c);
}

/* Setup callback to be called by cx88 common layer during device attach */
static int cx23880_video_setup(cx88_common_softc *arg)
{
   cx23880_video_softc *sc = (cx23880_video_softc*)arg;
   int error = cx88_i2c_attach(&(sc->i2c), 0, &(sc->common), sc->common.dev);
   if (!error)
   {
      error = cx88_video_setup(&(sc->video), CX88_VIDEO_STREAM_CX23880, &(sc->common),
         cx23880_video_i2c, sc->common.dev);
      if (error)
      {
         cx88_i2c_detach(&(sc->i2c));
      }
   }
   return error;
}

/* Cleanup callback to be called by cx88 common layer during device detach */
static int cx23880_video_teardown(cx88_common_softc *arg)
{
   cx23880_video_softc *sc = (cx23880_video_softc*)arg;
   cx88_video_teardown(&(sc->video));
   return cx88_i2c_detach(&(sc->i2c));
}

static int cx23880_video_start_capture(device_t dev, cx88_video_capture *cap)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   uint32_t input_format_reg = VIDEO_IFORM_VSYNC_ENABLE;
   uint32_t filter_control_reg = VIDEO_FILTER_CFILT;
   uint32_t output_format_reg = VIDEO_OFORM_FULL_LUMA | 
      cx88_write_field(VIDEO_OFORM_COMB_RANGE, 0x1F);
   if (sc->video.format.in_video_format == CX88_VIDEO_FMT_SECAM)
   {
      filter_control_reg |= VIDEO_FILTER_CLPF2;
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_CLPF1, VIDEO_CLPF1_29_TAP);
   }
   if (cap->in_input_type == CX88_VIDEO_INPUT_SVIDEO)
   {
      filter_control_reg |= VIDEO_FILTER_LNOTCH_DISABLE;
      input_format_reg |= (VIDEO_IFORM_SVID_CHROMA | VIDEO_IFORM_SVIDEO);
   }
   else if (cap->in_input_type == CX88_VIDEO_INPUT_VIP)
   {
      input_format_reg |= VIDEO_IFORM_VIP_INPUT;
   }
   if (sc->video.format.in_x_size < 640)
   {
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_COMB, VIDEO_COMB_CHROMA);
   }
   else
   {
      output_format_reg |= (VIDEO_OFORM_COMBALT | VIDEO_OFORM_PREV_REMOD);
      if (sc->video.format.in_video_format >= CX88_VIDEO_FMT_PALBDGHI)
      {
         output_format_reg |= VIDEO_OFORM_PAL_INV_PHASE;
      }
   }
   if (sc->video.format.in_even_field.enable && sc->video.format.in_odd_field.enable && 
      (sc->video.format.in_field_options & CX88_VIDEO_FIELD_INTERLACE))
   {
      filter_control_reg |= VIDEO_FILTER_VINT;
   }
   if (sc->video.format.in_x_size < 193)
   {
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_VFILT, VIDEO_VFILT_INTERP_5TAP);
   }
   else if (sc->video.format.in_x_size < 385)
   {
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_VFILT, VIDEO_VFILT_INTERP_3TAP);
   }
   if (cap->in_input_options & CX88_VIDEO_INPUT_AGC_ENABLE)
   {
      input_format_reg |= VIDEO_IFORM_AGC_ENABLE;
   }
   if (!(cap->in_input_options & CX88_VIDEO_INPUT_AGC_COARSE))
   {
      input_format_reg |= VIDEO_IFORM_NORMAL_AGC;
   }
   if (cap->in_input_options & CX88_VIDEO_INPUT_AGC_ADAPTIVE)
   {
      input_format_reg |= VIDEO_IFORM_ADAPTIVE_AGC;
   }
   if (cap->in_input_options & CX88_VIDEO_INPUT_COLORKILLER)
   {
      input_format_reg |= VIDEO_IFORM_COLOR_KILLER;
   }
   input_format_reg |= cx88_write_field(VIDEO_IFORM_INPUT_SOURCE, cap->in_input_source);
   input_format_reg |= cx88_write_field(VIDEO_IFORM_FORMAT, sc->video.format.in_video_format);
   cx88_write_reg(&(sc->common), VIDEO_REG_FILTER_EVEN, filter_control_reg);
   cx88_write_reg(&(sc->common), VIDEO_REG_FILTER_ODD, filter_control_reg);
   cx88_write_reg(&(sc->common), VIDEO_REG_OUTPUT_FORMAT, output_format_reg);
   cx88_write_reg(&(sc->common), VIDEO_REG_INPUT_FORMAT, input_format_reg);
   uint32_t sample_freq = sc->spec->sample_freq;
   uint32_t fsc8 = sample_freq;
   if (cap->in_input_type == CX88_VIDEO_INPUT_VIP)
   {
      sample_freq = 27000000;
   }
   uint64_t temp = ((uint64_t)28636363 * (1 << 17)) / sample_freq;
   cx88_write_reg(&(sc->common), VIDEO_REG_SAMPLE_RATE_CONV, (uint32_t) temp);
   if (sc->video.format.in_video_format == CX88_VIDEO_FMT_SECAM)
   {
      temp = ((uint64_t)34000000 * (1 << 22)) / sample_freq;
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP, (uint32_t) temp);
      temp = ((uint64_t)35250000 * (1 << 22)) / sample_freq;
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP_DR, (uint32_t) temp);
   }
   else
   {   
      temp = ((uint64_t)fsc8 * (1 << 22)) / sample_freq;
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP, (uint32_t) temp);
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP_DR, (uint32_t) temp);
   }  
   uint32_t agc_burst_delay = cx88_write_field(VIDEO_BURST_DELAY, 
      ((65 * sample_freq) / 20000000) + 21);
   agc_burst_delay |= cx88_write_field(VIDEO_AGC_DELAY, ((68 * sample_freq) / 20000000) + 15);
   cx88_write_reg(&(sc->common), VIDEO_REG_AGC_BURST_DELAY, agc_burst_delay);
   temp = ((uint64_t)(sc->spec->pixels_per_line) * sample_freq) / fsc8;
   uint32_t htotal_reg = cx88_write_field(VIDEO_HTOTAL, (uint32_t)temp);
   if (cap->in_input_type == CX88_VIDEO_INPUT_VIP)
   {
      if (sc->video.format.in_video_format >= CX88_VIDEO_FMT_PALBDGHI)
      {
         htotal_reg |= cx88_write_field(VIDEO_HLOTCH, VIDEO_HLNOTCH_135PAL);
      }
      else
      {
         htotal_reg |= cx88_write_field(VIDEO_HLOTCH, VIDEO_HLNOTCH_135NTSC);
      }
   }
   cx88_write_reg(&(sc->common), VIDEO_REG_HLNOTCH_HTOTAL, htotal_reg);
   temp = ((uint64_t)(sample_freq) * 16 * (1 << 20)) / 28636363;
   cx88_write_reg(&(sc->common), VIDEO_REG_PLL,
      cx88_write_field(VIDEO_PLL_DIV, (uint32_t)temp));
   uint16_t ticks = 0;
   int error = ETIMEDOUT;
   do
   {
      if (cx88_read_reg(&(sc->common), VIDEO_REG_STATUS) & VIDEO_STATUS_PLL_LOCK)
      {
         error = 0;
         break;  
      }
      pause("cx88p", 10);
      ticks += 10;
   }
   while ((ticks < cap->in_lock_timeout_ms) && (ticks < 60000));
   if (error)
   {
      device_printf(sc->common.dev, "Unable to acquire signal\n");
      return error;
   }
   pause("cx88p", cap->in_lock_delay_ms);
   if (!(cx88_read_reg(&(sc->common), VIDEO_REG_STATUS) & VIDEO_STATUS_PLL_LOCK))
   {
      device_printf(sc->common.dev, "Lost signal lock\n");
      return EIO;
   }
   /* Protect access to the shared PCI interrupt and ADC config registers. */
   cx88_lock_board(&(sc->common));
   if (cap->in_input_type == CX88_VIDEO_INPUT_SVIDEO)
   {
      cx88_write_reg(&(sc->common), CX23880_REG_ADC_CONFIG,
         cx88_read_reg(&(sc->common), CX23880_REG_ADC_CONFIG) | CX23880_CHROMA_VIDEO); 
   }
   else
   {
      cx88_write_reg(&(sc->common), CX23880_REG_ADC_CONFIG,
         cx88_read_reg(&(sc->common), CX23880_REG_ADC_CONFIG) & (~CX23880_CHROMA_VIDEO)); 
   }
   cx88_unlock_board(&(sc->common));
   uint32_t capture_control_reg = 0;
   if (sc->video.format.in_even_field.enable)
   {
      capture_control_reg |= VIDEO_CAPTURE_EVEN;
      if (sc->video.format.in_enable_vbi)
      {
         capture_control_reg |= VIDEO_CAPTURE_VBI_EVEN;
      }
   }
   if (sc->video.format.in_odd_field.enable)
   {
      capture_control_reg |= VIDEO_CAPTURE_ODD;
      if (sc->video.format.in_enable_vbi)
      {
         capture_control_reg |= VIDEO_CAPTURE_VBI_ODD;
      }
   }
   cx88_write_reg(&(sc->common), VIDEO_REG_CAPTURE_CONTROL, capture_control_reg); 
   return 0;
}

static void cx23880_video_stop_capture(device_t dev)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_CAPTURE_CONTROL, 0);
}

static int cx23880_video_set_format(device_t dev, cx88_video_format *fmt)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   uint8_t i = CX23880_FMT_TO_SPEC(fmt->in_video_format);
   if ((i >= CX23880_VIDEO_SPECS) || (fmt->in_pixel_format > CX88_VIDEO_NPIXFMTS))
   {
      return EINVAL;
   }
   sc->spec = &(specs[i]);
   if (fmt->in_enable_vbi)
   {
      cx88_write_reg(&(sc->common), VIDEO_REG_VBI_PACKET, 
         cx88_write_field(VIDEO_VBI_PACKET_SIZE, sc->spec->vbi_packets) |
         cx88_write_field(VIDEO_VBI_VDELAY, 10));
      fmt->out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size = (sc->spec->vbi_packets * 4) + 4;
      fmt->out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines = sc->spec->vdelay;
      if (!(fmt->in_odd_field.enable && fmt->in_even_field.enable))
      {
         fmt->out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines /= 2;
      }
   }
   uint32_t hscale = ((sc->spec->hactive * 4096) / fmt->in_cols) - 4096;
   uint32_t vscale = 0x10000 - ((512 * sc->spec->vactive) / fmt->in_rows) + 512;
   if ((fmt->in_cols > sc->spec->hactive) || (hscale > 0xFFFF) ||
       (fmt->in_rows > sc->spec->vactive) || (vscale <= 0xE000))
   {
      return EINVAL;
   } 
   vscale &= 0x1FFF;
   uint32_t hdelay = ((fmt->in_cols * sc->spec->hdelay) / sc->spec->hactive) & 0x3FE;
   hdelay += fmt->in_x_offset;
   uint32_t vdelay = sc->spec->vdelay + ((fmt->in_y_offset * sc->spec->vactive) / fmt->in_rows);
   uint32_t vactive = (fmt->in_y_size * sc->spec->vactive) / fmt->in_rows;
   if ((vdelay % 2) != 0)
   {
      vdelay -= 1;
   }
   cx88_write_reg(&(sc->common), VIDEO_REG_HSCALE_EVEN, hscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_HSCALE_ODD, hscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_VSCALE_EVEN, vscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_VSCALE_ODD, vscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_HDELAY_EVEN, hdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_HDELAY_ODD, hdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_EVEN, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_ODD, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_VIP_EVEN, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_VIP_ODD, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_HACTIVE_EVEN, fmt->in_x_size);
   cx88_write_reg(&(sc->common), VIDEO_REG_HACTIVE_ODD, fmt->in_x_size);
   cx88_write_reg(&(sc->common), VIDEO_REG_VACTIVE_EVEN, vactive);
   cx88_write_reg(&(sc->common), VIDEO_REG_VACTIVE_ODD, vactive);
   uint32_t color_format_reg = 0;
   if (fmt->in_even_field.enable)
   {
      if (fmt->in_even_field.swap & CX88_VIDEO_SWAP_WORDS)
      {
         color_format_reg |= VIDEO_COLOR_WSWAP_EVEN;
      }
      if (fmt->in_even_field.swap & CX88_VIDEO_SWAP_BYTES)
      {
         color_format_reg |= VIDEO_COLOR_BSWAP_EVEN;
      }
      color_format_reg |= cx88_write_field(VIDEO_COLOR_EVEN, fmt->in_pixel_format);
   }
   if (fmt->in_odd_field.enable)
   {
      if (fmt->in_odd_field.swap & CX88_VIDEO_SWAP_WORDS)
      {
         color_format_reg |= VIDEO_COLOR_WSWAP_ODD;
      }
      if (fmt->in_odd_field.swap & CX88_VIDEO_SWAP_BYTES)
      {
         color_format_reg |= VIDEO_COLOR_BSWAP_ODD;
      }
      color_format_reg |= cx88_write_field(VIDEO_COLOR_ODD, fmt->in_pixel_format);
   }
   cx88_write_reg(&(sc->common), VIDEO_REG_COLOR_FORMAT, color_format_reg);
   return 0;
}

static int cx23880_video_get_info(device_t dev, cx88_video_status *status)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   uint32_t status_reg = cx88_read_reg(&(sc->common), VIDEO_REG_STATUS);
   status->out_video_present = (status_reg & VIDEO_STATUS_VPRES) ? 1 : 0;
   status->out_play_mode = cx88_read_field(status_reg, VIDEO_STATUS_NSPLAY);
   status->out_horizontal_lock = (status_reg & VIDEO_STATUS_HLOCK) ? 1 : 0;
   status->out_system_type = cx88_read_field(status_reg, VIDEO_STATUS_NUML);
   return 0;
}

static int cx23880_video_set_hue(device_t dev, uint8_t hue)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_HUE, (uint32_t)hue);
   return 0;
}

static int cx23880_video_set_brightness(device_t dev, uint8_t brightness)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT,
      cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT), 
         VIDEO_BRIGHTNESS, (uint32_t)brightness));
   return 0;
}

static int cx23880_video_set_contrast(device_t dev, uint8_t contrast)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT,
      cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT), 
         VIDEO_CONTRAST, (uint32_t)contrast));
   return 0;
}

static int cx23880_video_set_sat_u(device_t dev, uint8_t sat)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_UV_SAT,
      cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_UV_SAT), 
         VIDEO_USAT, (uint32_t)sat));
   return 0;
}

static int cx23880_video_set_sat_v(device_t dev, uint8_t sat)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_UV_SAT,
      cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_UV_SAT), 
         VIDEO_VSAT, (uint32_t)sat));
   return 0;
}

static int cx23880_video_set_tdec(device_t dev, uint8_t fps)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_write_reg(&(sc->common), VIDEO_REG_TDEC,
      VIDEO_TDEC_CCLAMP_DISABLE | VIDEO_TDEC_EVEN_FIELD | cx88_write_field(VIDEO_TDEC_COUNT, (uint32_t)fps));
   return 0;
}

static int cx23880_video_filter(void *arg)
{
   cx23880_video_softc *sc = arg;
   return cx88_video_filter(&(sc->video));
}
   
static void cx23880_video_intr(void *arg)
{
   cx23880_video_softc *sc = arg;
   cx88_video_intr(&(sc->video));
}

/* Probe DEVMETHOD */
static int cx23880_video_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx23880_video_client);
}

/* Attach DEVMETHOD */
static int cx23880_video_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx23880_video_client);
}

/* Shutdown DEVMETHOD */
static int cx23880_video_shutdown(device_t dev)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   cx88_video_reset(&(sc->video), CX88_VIDEO_INVALID);
   return 0;
}

/* Suspend DEVMETHOD */
static int cx23880_video_suspend(device_t dev)
{
   cx23880_video_softc *sc = device_get_softc(dev);
   return cx88_video_suspend(&(sc->video));
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx23880_video_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_bus(&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx23880_video_get_i2c_dev(device_t dev, uint8_t addr)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_dev(&(sc->i2c), addr);
}

/* Resource reallocation DEVMETHOD */
static int cx23880_video_request_reconfig(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx23880_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return ENODEV;
   }
   return cx88_video_request_reconfig(&(sc->video));
}
 
static devclass_t cx23880_video_devclass;

/* DEVMETHODs */
static device_method_t cx23880_video_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,              cx23880_video_probe),
   DEVMETHOD(device_attach,             cx23880_video_attach),
   DEVMETHOD(device_detach,             cx88_common_detach),
   DEVMETHOD(device_shutdown,           cx23880_video_shutdown),
   DEVMETHOD(device_suspend,            cx23880_video_suspend),
   DEVMETHOD(device_resume,             cx88_common_resume),

   /* cx88 interface */
   CX88_METHODS(cx23880_video_request_reconfig),
             
   /* i2c host methods */
   CX88_I2C_METHODS(cx23880_video_get_i2c_bus, cx23880_video_get_i2c_dev),

   /* cx88 video interface */
   DEVMETHOD(cx88_video_start_capture,  cx23880_video_start_capture),
   DEVMETHOD(cx88_video_stop_capture,   cx23880_video_stop_capture),
   DEVMETHOD(cx88_video_set_format,     cx23880_video_set_format),
   DEVMETHOD(cx88_video_get_info,       cx23880_video_get_info),
   DEVMETHOD(cx88_video_set_hue,        cx23880_video_set_hue),
   DEVMETHOD(cx88_video_set_brightness, cx23880_video_set_brightness),
   DEVMETHOD(cx88_video_set_contrast,   cx23880_video_set_contrast),
   DEVMETHOD(cx88_video_set_sat_u,      cx23880_video_set_sat_u),
   DEVMETHOD(cx88_video_set_sat_v,      cx23880_video_set_sat_v),
   DEVMETHOD(cx88_video_set_tdec,       cx23880_video_set_tdec),

   {NULL, NULL}
};

/* Module declarations */
static driver_t cx23880_video_driver =
{
   "cx23880video",
   cx23880_video_methods,
   0
};

DRIVER_MODULE_ORDERED(cx23880video, pci, cx23880_video_driver, cx23880_video_devclass, NULL, NULL, SI_ORDER_ANY);
DRIVER_MODULE_ORDERED(cx23880video, cardbus, cx23880_video_driver, cx23880_video_devclass, NULL, NULL, SI_ORDER_ANY);

CX88_DRIVER_MODULE(cx23880video);
CX88_I2C_DRIVER_MODULE(cx23880video);
MODULE_DEPEND(cx23880mpeg, cx88videocore, CX88_VIDEO_MINVER, CX88_VIDEO_MODVER, CX88_VIDEO_MAXVER);
MODULE_VERSION(cx23880video, 1);

