/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "cx88_video.h"

#define MAX_VIDEO_BUFFER_SIZE (1 << 21)

static void cx88_video_reconfigure(cx88_video_softc *sc);
static int cx88_video_dmaprog(cx88_buffer *buffer, uint8_t bufindex);
static int cx88_video_configure(
   cx88_video_softc *sc,
   cx88_video_config *conf,
   struct thread *context);
static void cx88_video_release_channels(cx88_video_softc *sc);
static int cx88_video_geom(cx88_video_softc *sc, cx88_video_format *fmt);
static int cx88_video_set_clip(cx88_video_softc *sc, cx88_video_clip *clip);
static int cx88_video_start_capture(cx88_video_softc *sc, cx88_video_capture *cap);
static int cx88_video_transfer(cx88_video_softc *sc, cx88_video_wait *wait, boolean_t nonblocking);
static int cx88_video_lineprog(
   cx88_buffer *buffer,
   bus_size_t offset,
   cx88_video_clip_pattern *pattern);
static int cx88_video_fieldprog(
   cx88_buffer *buffer,
   cx88_video_channel_config *config,
   uint16_t *line,
   bus_size_t *offset,
   bus_size_t max_offset,
   bus_size_t line_size,
   uint8_t interlace);

static d_ioctl_t cx88_video_ioctl;
static d_poll_t cx88_video_poll;

static struct cdevsw cx88_video_cdevsw =
{
   .d_version = D_VERSION,
   .d_name = "cx88video",
   .d_ioctl = cx88_video_ioctl, /* ioctl */
   .d_poll = cx88_video_poll,   /* poll */
};

static struct unrhdr *cx88_video_units = NULL;

static const cx88_video_stream_desc cx88_video_streams[] = 
{
   {
      .stream = CX88_VIDEO_STREAM_CX23880,
      .dma_channels =
      {
         CX88_CH_VIDEO_PACKED,
         CX88_CH_VIDEO_U,
         CX88_CH_VIDEO_V,
         CX88_CH_VIDEO_VBI
      },
      .reg_gp_count =
      {
         CX23880_VIDEO_REG_Y_GP_COUNT,
         CX23880_VIDEO_REG_U_GP_COUNT,
         CX23880_VIDEO_REG_V_GP_COUNT,
         CX23880_VIDEO_REG_VBI_GP_COUNT
      },
      .reg_dma_control = CX23880_VIDEO_REG_DMA_CONTROL,
      .dma_control_mask =
      {
         VIDEO_Y_FIFO_ENABLE | VIDEO_Y_RISC_ENABLE,
         CX23880_VIDEO_U_FIFO_ENABLE | CX23880_VIDEO_U_RISC_ENABLE,
         CX23880_VIDEO_V_FIFO_ENABLE | CX23880_VIDEO_V_RISC_ENABLE,
         CX23880_VIDEO_VBI_FIFO_ENABLE | CX23880_VIDEO_VBI_RISC_ENABLE
      },
      .reg_pci_intr_mask = CX23880_REG_PCI_INTMASK,
      .pci_intr_mask = CX23880_PCI_INT_VIDEO,
      .reg_intr_mask = CX23880_VIDEO_REG_INT_MASK,
      .reg_intr_stat = CX23880_VIDEO_REG_INT_STAT,
      .reg_intr_mstat = CX23880_VIDEO_REG_INT_MSTAT,
      .risci1_mask =
      {
         VIDEO_INT_Y_RISCI1,
         CX23880_VIDEO_INT_U_RISCI1,
         CX23880_VIDEO_INT_V_RISCI1,
         CX23880_VIDEO_INT_VBI_RISCI1
      },
      .risci2_mask =
      {
         VIDEO_INT_Y_RISCI2,
         CX23880_VIDEO_INT_U_RISCI2,
         CX23880_VIDEO_INT_V_RISCI2,
         CX23880_VIDEO_INT_VBI_RISCI2
      },
      .sync_err_mask =
      {
         VIDEO_INT_Y_SYNC,
         CX23880_VIDEO_INT_U_SYNC,
         CX23880_VIDEO_INT_V_SYNC,
         CX23880_VIDEO_INT_VBI_SYNC
      },
      .overflow_mask =
      {
         VIDEO_INT_Y_OVERFLOW,
         CX23880_VIDEO_INT_U_OVERFLOW,
         CX23880_VIDEO_INT_V_OVERFLOW,
         CX23880_VIDEO_INT_VBI_OVERFLOW
      },
      .opcode_err_mask =
      {
         VIDEO_INT_OPCODE_ERR,
         0,
         0,
         0
      },
      .bus_err_mask = CX23880_VIDEO_INT_PCI_ABORT | CX23880_VIDEO_INT_PAR_ERR |
         CX23880_VIDEO_INT_RIP_ERR,
      .reg_devcntl2 = CX23880_REG_DEVCNTL2,
      .pixformats = (1 << CX88_VIDEO_NPIXFMTS) - 1 
   },
   {
      .stream = CX88_VIDEO_STREAM_CX23885_VIDA,
      .dma_channels =
      {
         CX88_CH_VIDEO_PACKED,
         CX88_CH_VIDEO_U,
         CX88_CH_VIDEO_V,
         CX88_CH_VIDEO_VBI
      },
      .reg_gp_count =
      {
         CX23885_VIDA_REG_GP_COUNT,
         0xFFFFFFFF,
         0xFFFFFFFF,
         CX23885_VIDA_REG_VBI_GP_COUNT
      },
      .reg_dma_control = CX23885_VIDA_REG_DMA_CONTROL,
      .dma_control_mask =
      {
         VIDEO_Y_FIFO_ENABLE | VIDEO_Y_RISC_ENABLE,
         0,
         0,
         CX23885_VIDEO_VBI_FIFO_ENABLE | CX23885_VIDEO_VBI_RISC_ENABLE
      },
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_VIDEO_A,
      .reg_intr_mask = CX23885_VIDEO_A_REG_INT_MASK,
      .reg_intr_stat = CX23885_VIDEO_A_REG_INT_STAT,
      .reg_intr_mstat = CX23885_VIDEO_A_REG_INT_MSTAT,
      .risci1_mask =
      {
         VIDEO_INT_Y_RISCI1,
         0,
         0,
         CX23885_VIDEO_INT_VBI_RISCI1,
      },
      .risci2_mask =
      {
         VIDEO_INT_Y_RISCI2,
         0,
         0,
         CX23885_VIDEO_INT_VBI_RISCI2,
      },
      .sync_err_mask =
      {
         VIDEO_INT_Y_SYNC,
         0,
         0,
         CX23885_VIDEO_INT_VBI_SYNC
      },
      .overflow_mask =
      {
         VIDEO_INT_Y_OVERFLOW,
         0,
         0,
         CX23885_VIDEO_INT_VBI_OVERFLOW
      },
      .opcode_err_mask =
      {
         VIDEO_INT_OPCODE_ERR,
         0,
         0,
         CX23885_VIDEO_INT_VBI_OPCODE_ERR
      },
      .reg_devcntl2 = CX23885_REG_DEVCNTL2,
      .pixformats = (1 << CX88_VIDEO_YUV422)
   },
   {
      .stream = CX88_VIDEO_STREAM_CX23885_VIDB,
      .dma_channels =
      {
         CX88_CH_VIDEO_PACKED_2,
         CX88_CH_VIDEO_U,
         CX88_CH_VIDEO_V,
         CX88_CH_VIDEO_VBI_2
      },
      .reg_gp_count =
      {
         CX23885_VIDA_REG_GP_COUNT,
         0xFFFFFFFF,
         0xFFFFFFFF,
         CX23885_VIDA_REG_VBI_GP_COUNT
      },
      .reg_dma_control = CX23885_VIDB_REG_DMA_CONTROL,
      .dma_control_mask =
      {
         VIDEO_Y_FIFO_ENABLE | VIDEO_Y_RISC_ENABLE,
         0,
         0,
         CX23885_VIDEO_VBI_FIFO_ENABLE | CX23885_VIDEO_VBI_RISC_ENABLE
      },
      .reg_pci_intr_mask = CX23885_REG_PCI_INTMASK,
      .pci_intr_mask = CX23885_PCI_INT_VIDEO_B,
      .reg_intr_mask = CX23885_VIDEO_B_REG_INT_MASK,
      .reg_intr_stat = CX23885_VIDEO_B_REG_INT_STAT,
      .reg_intr_mstat = CX23885_VIDEO_B_REG_INT_MSTAT,
      .risci1_mask =
      {
         VIDEO_INT_Y_RISCI1,
         0,
         0,
         CX23885_VIDEO_INT_VBI_RISCI1,
      },
      .risci2_mask =
      {
         VIDEO_INT_Y_RISCI2,
         0,
         0,
         CX23885_VIDEO_INT_VBI_RISCI2,
      },
      .sync_err_mask =
      {
         VIDEO_INT_Y_SYNC,
         0,
         0,
         CX23885_VIDEO_INT_VBI_SYNC
      },
      .overflow_mask =
      {
         VIDEO_INT_Y_OVERFLOW,
         0,
         0,
         CX23885_VIDEO_INT_VBI_OVERFLOW
      },
      .opcode_err_mask =
      {
         VIDEO_INT_OPCODE_ERR,
         0,
         0,
         CX23885_VIDEO_INT_VBI_OPCODE_ERR
      },
      .reg_devcntl2 = CX23885_REG_DEVCNTL2,
      .pixformats = (1 << CX88_VIDEO_YUV422)
   }
};

int cx88_video_setup(
   cx88_video_softc *sc,
   cx88_video_stream_t stream,
   cx88_common_softc *common,
   cx88_i2c_softc* (*i2c_accessor)(cx88_common_softc*, uint8_t),
   device_t videocore_dev)
{
   int error = 0;
   if (stream >= CX88_VIDEO_NUM_STREAMS)
   {
      return EINVAL;
   }
   sc->stream = &(cx88_video_streams[stream]);
   sc->common = common;
   sc->i2c_accessor = i2c_accessor;
   sc->videocore_dev = videocore_dev;
   sc->reconfig_requested = FALSE;
   sc->in_use = FALSE;
   sc->unit = -1;
   sc->devnode = NULL;
   sc->dma_program_valid = TRUE;
   mtx_init(&(sc->isrlock), "CX88 Video ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 Video SX lock");
   uint8_t i;   
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      sc->channel_configs[i].x_header = 0;
      sc->channel_configs[i].x_trailer = 0;
      sc->channel_configs[i].clip_patterns = NULL;
      sc->channel_configs[i].clip_pattern_size = 0;
      sc->channel_configs[i].last_sync_buffer = 0;
      sc->channel_configs[i].even_frame_available = 0;
      sc->channel_configs[i].odd_frame_available = 0;
      sc->channel_configs[i].enabled = TRUE;
   }
   int disabled = 0;
   if ((sc->stream->stream != CX88_VIDEO_STREAM_CX23880) ||
       ((resource_int_value(device_get_name(sc->common->dev), device_get_unit(sc->common->dev),
         "noplanar", &disabled) == 0) && disabled))
   {
      sc->channel_configs[CX88_VIDEO_CHANNEL_U].enabled = FALSE;
      sc->channel_configs[CX88_VIDEO_CHANNEL_V].enabled = FALSE;
   }
   disabled = 0;
   if ((resource_int_value(device_get_name(sc->common->dev), device_get_unit(sc->common->dev),
       "novbi", &disabled) == 0) && disabled)
   {
      sc->channel_configs[CX88_VIDEO_CHANNEL_VBI].enabled = FALSE;
   }
   for (i = 0; !error && (i < CX88_VIDEO_CHANNELS); ++i)
   {
      if (sc->channel_configs[i].enabled)
      {
         error = cx88_init_channel(&(sc->channels[i]), sc->common,
			   sc->stream->dma_channels[i], MAX_VIDEO_BUFFER_SIZE);
      }
   }
   if (!error && ((sc->unit = alloc_unr(cx88_video_units)) == -1))
   {
      error = ENOMEM;
   }
   if (!error)
   {
      /* Create device file--this should be done as the last step in initialization */
      sc->devnode = make_dev(&cx88_video_cdevsw, sc->unit,
         UID_ROOT, GID_WHEEL, 0644, "cx88video%d", sc->unit);
      if (sc->devnode == NULL)
      {
         error = ENOMEM;
      }
      else
      {
         sc->devnode->si_drv1 = sc;
         cx88_video_reset(sc, CX88_VIDEO_UNCONFIGURED);
         /* Protect access to shared PCI interrupt mask register */
         cx88_lock_board(sc->common);
         cx88_write_reg(sc->common, sc->stream->reg_pci_intr_mask,
            (cx88_read_reg(sc->common, sc->stream->reg_pci_intr_mask) | sc->stream->pci_intr_mask));
         cx88_unlock_board(sc->common);
      }
   }
   if (error)
   {
      if (sc->unit != -1)
      {
         free_unr(cx88_video_units, sc->unit);
      }
      for (uint8_t j = 0; j < i; ++j)
      {
         if (sc->channel_configs[j].enabled)
         {
            cx88_destroy_channel (&(sc->channels[j]));
         }
      }
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   }
   return error;
}

void cx88_video_teardown(cx88_video_softc *sc)
{
   if (sc->devnode == NULL)
   {
      return;
   }
   uint8_t i;
   sx_slock(&(sc->sxlock));
   cx88_video_reset(sc, CX88_VIDEO_INVALID);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   /* Remove user-configurable data--buffers, channel allocation, etc. */
   cx88_video_release_channels(sc);
   sx_xunlock(&(sc->sxlock));
   destroy_dev(sc->devnode);
   sc->devnode = NULL;
   free_unr(cx88_video_units, sc->unit);
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      if (sc->channel_configs[i].enabled)
      {
         cx88_destroy_channel (&(sc->channels[i]));
      }
      free(sc->channel_configs[i].clip_patterns, M_CX88);
   }
   mtx_destroy(&(sc->isrlock));
   sx_destroy(&(sc->sxlock));
}

void cx88_video_reset(cx88_video_softc *sc, cx88_video_state_t state)
{
   mtx_lock(&(sc->isrlock));
	sc->intr_mask = 0;
   /* Stop any data transfer */
   cx88_write_reg(sc->common, sc->stream->reg_dma_control, 0);
   /* Disable interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask, 0);
   /* Disable video capture */
   CX88_VIDEO_STOP_CAPTURE(sc->videocore_dev);
   cx88_read_reg(sc->common, sc->stream->reg_intr_mask); /* Flush posted writes */

   /* Reset capture status values */
   sc->channels_avail = 0;
   sc->isr_status = CX88_VIDEO_WAIT_ABORTED;

   /* Our state value should never be higher than CX88_VIDEO_CONFIGURED, as we won't be capturing once reset is done */
   sc->state = CX88_VIDEO_CONFIGURED;
   if (state < sc->state)
   {
      /* Reset alone should never raise the state value */
      sc->state = state;  
   }
   
   wakeup(&(sc->channels_avail));
   selwakeup(&(sc->selectors));
   mtx_unlock(&(sc->isrlock));
}

static int cx88_video_transfer(cx88_video_softc *sc, cx88_video_wait *wait, boolean_t nonblocking)
{
   int error = 0;
   mtx_lock(&(sc->isrlock));
   if (sc->state != CX88_VIDEO_CAPTURING)
   {
      mtx_unlock(&(sc->isrlock));
      return ENXIO;
   }
   if (sc->channel_mask & wait->in_channel_mask)
   {
      device_printf(sc->common->dev, "Wait already in progress for conflicting channel(s)\n");
      mtx_unlock(&(sc->isrlock));
      return EINVAL;
   }
   sc->channel_mask |= wait->in_channel_mask;
   while (!error &&
          !(sc->isr_status & CX88_VIDEO_CAP_FATAL) &&
          !(sc->channels_avail & wait->in_channel_mask))
   {
      if (!nonblocking)
      {
         error = msleep(&(sc->channels_avail), &(sc->isrlock), PCATCH, 
            "cx88vi", wait->in_timeout_ms);
      }
      else
      {
         error = EWOULDBLOCK;
      }
   }
   sc->channel_mask &= (~(wait->in_channel_mask));
   wait->out_status = sc->isr_status;
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
		if (!error && (sc->channels_avail & (CX88_VIDEO_FIELD_EVEN(i) | CX88_VIDEO_FIELD_ODD(i))))
		{
			/* Perform any necessary cache/bounce buffer synchronization.
			 * Do this in a loop to account for the unlikely case where
			 * we've missed an interrupt and multiple buffers need to be 
			 * synchronized */
			uint8_t sync_bufindex = sc->channel_configs[i].last_sync_buffer;
			uint8_t last_bufindex = (sc->format.in_odd_field.enable ? 
				sc->channel_configs[i].odd_frame_available : 
				sc->channel_configs[i].even_frame_available);
			do
			{
				if (sync_bufindex >= sc->channels[i].num_buffers)
				{
					sync_bufindex = 0;
				}
				cx88_buffer *buffer = &(sc->channels[i].buffers[sync_bufindex]);
				/*Synchronize the buffer */
				cx88_mark_transfer_complete(buffer);
				cx88_prepare_for_transfer(buffer);
			}
			while ((++sync_bufindex) != last_bufindex);
			sc->channel_configs[i].last_sync_buffer = (sc->format.in_even_field.enable ? 
				sc->channel_configs[i].even_frame_available : 
				sc->channel_configs[i].odd_frame_available);
		}
      wait->channels[i].out_even_frame_available = sc->channel_configs[i].even_frame_available;
      wait->channels[i].out_odd_frame_available = sc->channel_configs[i].odd_frame_available;
      if (cx88_read_reg(sc->common, sc->stream->reg_intr_stat) & sc->stream->overflow_mask[i])
      {
         wait->channels[i].overflow = 1;
      }
      else
      {
         wait->channels[i].overflow = 0;  
      }
   }
   if (!error)
	{
      sc->channels_avail &= (~(sc->channels_avail & wait->in_channel_mask));
   }
   sc->isr_status = CX88_VIDEO_CAP_NOERR;
   mtx_unlock(&(sc->isrlock));
   /* For blocking I/O, msleep() timed out */
   if ((error == EWOULDBLOCK) && wait->in_timeout_ms)
   {
      wait->out_status |= CX88_VIDEO_CAP_TIMEOUT;
      error = 0;
   }
   else if (error)
   {
      wait->out_status |= CX88_VIDEO_WAIT_ABORTED;
   }
   return error;
}

static int cx88_video_start_capture(cx88_video_softc *sc, cx88_video_capture *cap)
{
   cx88_video_reset(sc, sc->state);
   if (sc->state != CX88_VIDEO_CONFIGURED)
   {
      return ENXIO;
   }
   int error = 0;
   sc->isr_status = CX88_VIDEO_CAP_NOERR;
   uint32_t dmac_reg = 0;
   sc->intr_mask = sc->stream->bus_err_mask; 
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
		sc->channel_configs[i].last_sync_buffer = 0;
      sc->channel_configs[i].even_frame_available = 0;
      sc->channel_configs[i].odd_frame_available = 0;
      if (sc->format.out_geoms[i].line_size != 0)
      {
         if ((error = cx88_start_channel(&(sc->channels[i]))))
         {
            return error;
         }
         sc->intr_mask |= (sc->stream->overflow_mask[i] | sc->stream->sync_err_mask[i]);
         if (!(cap->in_no_frame_intrs))
         {
            sc->intr_mask |= (sc->stream->risci1_mask[i] | sc->stream->risci2_mask[i]);
         }
         dmac_reg |= (sc->stream->opcode_err_mask[i] | sc->stream->dma_control_mask[i]);
      }
   }
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(sc->common, sc->stream->reg_intr_stat,
      cx88_read_reg(sc->common, sc->stream->reg_intr_stat));
   if ((error = CX88_VIDEO_START_CAPTURE(sc->videocore_dev, cap)))
   {
      return error;
   }
   sc->state = CX88_VIDEO_CAPTURING;
   cx88_write_reg(sc->common, sc->stream->reg_intr_mask, sc->intr_mask);
   cx88_write_reg(sc->common, sc->stream->reg_devcntl2, CX88_RUN_RISC);
   cx88_write_reg(sc->common, sc->stream->reg_dma_control, dmac_reg);
   return 0;
}

/* Configures channels and buffers */
static int cx88_video_configure(
   cx88_video_softc *sc,
   cx88_video_config *config,
   struct thread *context)
{
   int error = 0;
   uint8_t i;
   if (sc->state < CX88_VIDEO_FORMATTED)
   {
      return ENXIO;
   }
   cx88_video_reset(sc, CX88_VIDEO_FORMATTED); /* Ensure there is no outstanding capture */
   /* Validate ioctl parameters */
   if ((config->in_num_frames == 0) || (config->in_num_frames > CX88_VIDEO_MAX_FRAMES))
   {
      return EINVAL;
   }
   DIAGNOSTIC(device_printf(sc->common->dev, "Configuring video channels\n"))
   uint8_t old_num_frames = sc->num_frames;
   sc->num_frames = (config->in_one_shot ? config->in_num_frames : 0);
   for (i = 0; (i < CX88_VIDEO_CHANNELS) && !error; ++i)
   {
      if (sc->format.out_geoms[i].line_size != 0)
      {
         boolean_t dma_program_valid = sc->dma_program_valid;
         if ((config->bufconfig[i].in_x_header != sc->channel_configs[i].x_header) ||
             (config->bufconfig[i].in_x_trailer != sc->channel_configs[i].x_trailer))
         {
            sc->channel_configs[i].x_header = config->bufconfig[i].in_x_header;
            sc->channel_configs[i].x_trailer = config->bufconfig[i].in_x_trailer;
            dma_program_valid = FALSE;
         }
         size_t line_size = sc->format.out_geoms[i].line_size +
            (sc->channel_configs[i].x_header + sc->channel_configs[i].x_trailer);
         size_t buffer_size = line_size * sc->format.out_geoms[i].num_lines;
         /* Allocate the channel */
         error = cx88_configure_channel(
            &(sc->channels[i]),
            0,
            sc->format.out_geoms[i].line_size,
            sc->format.out_geoms[i].line_size,
            0);
         if (!error)
         {      
            /* Request the buffer structures */
            error = cx88_request_buffers (&(sc->channels[i]), config->in_num_frames, 0);
         }
         uint8_t j;
         for (j = 0; (j < config->in_num_frames) && !error; ++j)
         {
            if (config->bufconfig[i].in_buffers[j] == NULL)
            {
               error = EINVAL;
               break;  
            }
            /* Each cx88 buffer is a segment of the user-allocated buffer */
            void *bufptr = (uint8_t *)(config->bufconfig[i].in_buffers[j]) + 
               (line_size * config->bufconfig[i].in_y_header);
            if ((old_num_frames != sc->num_frames) && 
                (((old_num_frames != 0) && (j == (old_num_frames - 1))) ||
                 ((sc->num_frames != 0) && (j == (sc->num_frames - 1)))))
            {
               dma_program_valid = FALSE;   
            }
            if (!dma_program_valid)
            {
               cx88_invalidate_dma_program(&(sc->channels[i].buffers[j]));
            }
            /* Configure each buffer */
            error = cx88_configure_buffer(
               &(sc->channels[i].buffers[j]),
               bufptr, /* This will be NULL for kernel-allocated buffers */
               context,
               buffer_size,
               ((j == (config->in_num_frames - 1)) ? 0 : (j + 1)),
               cx88_video_dmaprog,
               sc);
         }
         if (!error)
         {
            /* Now commit all changes we've made to the channel configuration */
            error = cx88_commit_channel(&(sc->channels[i]));
         }
      }
      else
      {      
         cx88_release_channel (&(sc->channels[i])); /* Release all channel and buffer resources */
      }
   }
   if (!error)
   {
      sc->dma_program_valid = TRUE;
      sc->state = CX88_VIDEO_CONFIGURED;
   }
   else
   {
      cx88_video_release_channels(sc);
   }
   return error;
}

static void cx88_video_release_channels(cx88_video_softc *sc)
{
   uint8_t i;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      cx88_release_channel (&(sc->channels[i])); /* Release all channel and buffer resources */
   }
}

static void cx88_video_dtor(void *data)
{
   cx88_video_softc *sc = data;
   if (sc == NULL) return;
   DIAGNOSTIC(device_printf(sc->common->dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   cx88_video_reset(sc, sc->state);
   uint8_t i;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      /* If we're using user-allocated buffers, release them as
      * they can't be persisted across open() and close() in
      * different processes */
      cx88_free_userbufs(&(sc->channels[i]), NULL);
   }
   if (sc->reconfig_requested)
   {
      cx88_video_reconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
   sc->in_use = FALSE;
   sx_xunlock(&(sc->sxlock));
}

/* Establish video frame geometry and pixel format */
static int cx88_video_geom(cx88_video_softc *sc, cx88_video_format *fmt)
{
   if (sc->state < CX88_VIDEO_UNCONFIGURED)
   {
      return ENXIO;
   }
   if ((!(fmt->in_even_field.enable) && !(fmt->in_odd_field.enable)) ||
       (fmt->in_field_options & ~(CX88_VIDEO_FIELD_INTERLACE | CX88_VIDEO_FIELD_EVEN_FIRST)) ||
       (fmt->in_rows == 0) || (fmt->in_cols == 0) ||
       (fmt->in_x_size == 0) || (fmt->in_y_size == 0) ||
       ((fmt->in_x_offset + fmt->in_x_size) > fmt->in_cols) ||
       ((fmt->in_y_offset + fmt->in_y_size) > fmt->in_rows))
   {
      return EINVAL;
   }
   for (uint8_t i = CX88_VIDEO_CHANNEL_U; i < CX88_VIDEO_CHANNELS; ++i)
   {
      fmt->out_geoms[i].line_size = 0;
      fmt->out_geoms[i].num_lines = 0;
   }
   uint16_t num_lines = fmt->in_y_size;
   if (!(fmt->in_odd_field.enable && fmt->in_even_field.enable))
   {
      if (!(fmt->in_field_options & CX88_VIDEO_FIELD_INTERLACE))
      {
         num_lines /= 2;
      }
   }
   fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines = num_lines;
   switch (fmt->in_pixel_format)
   {
      case CX88_VIDEO_RGB32:
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = 4 * fmt->in_x_size;
         break;
      case CX88_VIDEO_BGR24:
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = 3 * fmt->in_x_size;
         break;
      case CX88_VIDEO_RGB16:
      case CX88_VIDEO_RGB15:
      case CX88_VIDEO_YUV422:
         if (fmt->in_x_size % 2)
         {
            return EINVAL;  
         }
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = 2 * fmt->in_x_size;
         break;
      case CX88_VIDEO_BTYUV411:
         if (fmt->in_x_size % 4)
         {
            return EINVAL;  
         }
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = (3 * fmt->in_x_size) / 2;
         break;
      case CX88_VIDEO_Y8:
      case CX88_VIDEO_TGB8:
      case CX88_VIDEO_YCC422:
      case CX88_VIDEO_YCC411:
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = fmt->in_x_size;
         break;
      default:
         return EINVAL;
   }
   if ((fmt->in_pixel_format == CX88_VIDEO_YCC422) || (fmt->in_pixel_format == CX88_VIDEO_YCC411))
   {
      if (!(sc->channel_configs[CX88_VIDEO_CHANNEL_U].enabled) ||
          !(sc->channel_configs[CX88_VIDEO_CHANNEL_V].enabled))
      {
         device_printf(sc->common->dev, "Planar capture unavailable or disabled by device hint\n");
         return ENXIO;
      }
      fmt->out_geoms[CX88_VIDEO_CHANNEL_U].num_lines = num_lines;
      fmt->out_geoms[CX88_VIDEO_CHANNEL_V].num_lines = num_lines;
      uint8_t scale_factor = ((fmt->in_pixel_format == CX88_VIDEO_YCC422) ? 2 : 4);
      if (fmt->in_x_size % scale_factor)
      {
         return EINVAL;
      }
      fmt->out_geoms[CX88_VIDEO_CHANNEL_U].line_size = fmt->in_x_size / scale_factor;
      fmt->out_geoms[CX88_VIDEO_CHANNEL_V].line_size = fmt->in_x_size / scale_factor;  
   }
   if ((fmt->in_enable_vbi) && (!(sc->channel_configs[CX88_VIDEO_CHANNEL_VBI].enabled)))
   {
      device_printf(sc->common->dev, "VBI capture disabled by device hint\n");
      return ENXIO;
   }
   int error = 0;
   if ((error = CX88_VIDEO_SET_FORMAT(sc->videocore_dev, fmt)))
   {
      return error;
   }
   if ((sc->state < CX88_VIDEO_FORMATTED) || 
       (sc->format.in_field_options != fmt->in_field_options) ||
       (sc->format.in_odd_field.enable != fmt->in_odd_field.enable) ||
       (sc->format.in_even_field.enable != fmt->in_even_field.enable))
   {
      sc->dma_program_valid = FALSE;
   }
   sc->format = *fmt;
   cx88_video_reset(sc, CX88_VIDEO_FORMATTED);
   return 0;
}

static int cx88_video_set_clip(cx88_video_softc *sc, cx88_video_clip *clip)
{
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      if (clip->channels[i].in_pattern_size == 0)
      {
         if (sc->channel_configs[i].clip_pattern_size != 0)
         {
            free(sc->channel_configs[i].clip_patterns, M_CX88);
            sc->channel_configs[i].clip_patterns = NULL;
            sc->channel_configs[i].clip_pattern_size = 0;
            sc->dma_program_valid = FALSE;  
         }
         continue;
      }
      else if ((clip->channels[i].in_patterns == NULL) || (clip->channels[i].in_pattern_size > CX88_VIDEO_MAX_PATTERN_SIZE))
      {
         return EINVAL;  
      }
      void *patterns = realloc(sc->channel_configs[i].clip_patterns,
         clip->channels[i].in_pattern_size, M_CX88, M_WAITOK);
      if (patterns == NULL)
      {
         return ENOMEM;
      }
      sc->channel_configs[i].clip_patterns = patterns;
      sc->channel_configs[i].clip_pattern_size = clip->channels[i].in_pattern_size;
      sc->dma_program_valid = FALSE;
      int error = copyin(clip->channels[i].in_patterns, sc->channel_configs[i].clip_patterns,
         sc->channel_configs[i].clip_pattern_size);
      if (error)
      {
         return error;
      }
   }
   return 0;
}

/* Ioctl file operation */
static int cx88_video_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   cx88_video_softc *sc = dev->si_drv1;
   int error = 0;
   if (sc == NULL)
   {
      return ENXIO;
   }
   void *cdp = NULL;
   switch (cmd)
   {
      case CX88_COMMON_CARD_ID: /* Retrieve vendor-specific identifiers */
         *((uint32_t*)data) = cx88_get_card(sc->common->dev);
         break;
      case CX88_VIDEO_FORMAT: /* Establish video geometry */
         sx_xlock(&(sc->sxlock));
         if (!(sc->in_use) || ((error = devfs_get_cdevpriv(&cdp)) == 0))
         {
            error = cx88_video_geom(sc, (cx88_video_format*)data);
            if (!error && (cdp == NULL))
            {
               error = devfs_set_cdevpriv(sc, cx88_video_dtor);
            }
            if (!error) sc->in_use = TRUE;
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_CONFIG: /* Configure video buffers */
         sx_xlock(&(sc->sxlock));
         if ((error = devfs_get_cdevpriv(&cdp)) == 0)
         {
            error = cx88_video_configure(sc, (cx88_video_config*)data, td);
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_I2C_CONFIG: /* Configure I2C devices */
      {
         cx88_common_i2c_config *conf = (cx88_common_i2c_config*)data;
         cx88_i2c_softc *i2c_sc = sc->i2c_accessor(sc->common, conf->in_bus_number);
         if (i2c_sc == NULL)
         {
            error = EINVAL;
         }
         else
         {
            error = cx88_i2c_update_devices(i2c_sc, conf);
         }
         break;
      }
      case CX88_VIDEO_CAPTURE: /* Start capturing data */
         sx_xlock(&(sc->sxlock));
         if ((error = devfs_get_cdevpriv(&cdp)) == 0)
         {
            error = cx88_video_start_capture(sc, (cx88_video_capture*)data);
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_STOP: /* Stop capturing data */
         /* Wait and stop are slock'ed so that a stop can be issued with a pending wait */
         sx_slock(&(sc->sxlock));
         if ((error = devfs_get_cdevpriv(&cdp)) == 0)
         {
            cx88_video_reset(sc, sc->state);
         }
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_WAIT: /* Wait for data to become available */
         sx_slock(&(sc->sxlock));
         if ((error = devfs_get_cdevpriv(&cdp)) == 0)
         {
            error = cx88_video_transfer(sc, (cx88_video_wait*)data, ((flags & O_NONBLOCK) ? TRUE : FALSE));
         }
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_GPIO: /* Modify GPIO pin state */
         cx88_gpioctl(sc->common, (cx88_common_gpio*)data);
         break;
      case CX88_COMMON_PERIPH_RESET:
         cx88_peripheral_reset(sc->common, *((uint32_t*)data));
         break;
      case CX88_VIDEO_HUE:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_SET_HUE(sc->videocore_dev, *((uint8_t*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_BRIGHTNESS:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_SET_BRIGHTNESS(sc->videocore_dev, *((uint8_t*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_CONTRAST:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_SET_CONTRAST(sc->videocore_dev, *((uint8_t*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_SAT_U:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_SET_SAT_U(sc->videocore_dev, *((uint8_t*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_SAT_V:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_SET_SAT_V(sc->videocore_dev, *((uint8_t*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_DROP_FRAMES:
         sx_xlock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_SET_TDEC(sc->videocore_dev, *((uint8_t*)data)));
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_GET_STATUS:
         sx_slock(&(sc->sxlock));
         error = ((sc->state < CX88_VIDEO_UNCONFIGURED) ?
            ENXIO : CX88_VIDEO_GET_INFO(sc->videocore_dev, (cx88_video_status*)data));
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_CLIP:
         sx_xlock(&(sc->sxlock));
         error = cx88_video_set_clip(sc, (cx88_video_clip*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX23885_COMMON_MC_GPIO:
         error = cx23885_mc_gpioctl(sc->common, (cx88_common_gpio*)data);
         break;
      case CX23885_COMMON_BRIDGE:
         snprintf((char*)data, CX88_COMMON_MAX_DRIVER_NAME, "%s", device_get_nameunit(sc->common->dev));
         break;
      case CX23885_COMMON_GET_STREAMS:
      {
         uint32_t *stream = (uint32_t*)data;
         switch(sc->stream->stream)
         {
            case CX88_VIDEO_STREAM_CX23885_VIDA:
               *stream = CX23885_STREAM_VIDEO1;
               break;
            case CX88_VIDEO_STREAM_CX23885_VIDB:
               *stream = CX23885_STREAM_VIDEO2;
               break;
            default:
               *stream = CX23885_STREAM_NONE;
               break;
         }
         break;
      }
      case CX88_VIDEO_PIXFORMATS:
      {
         uint32_t pixfmts = sc->stream->pixformats;
         if (!sc->channel_configs[CX88_VIDEO_CHANNEL_U].enabled ||
             !sc->channel_configs[CX88_VIDEO_CHANNEL_V].enabled)
         {
            pixfmts &= ~(CX88_VIDEO_YCC422 | CX88_VIDEO_YCC411);
         }
         *((uint32_t*)data) = pixfmts;
         break;
      }
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

/* Select/poll file operation */
static int cx88_video_poll(struct cdev *dev, int events, struct thread *td)
{   
   cx88_video_softc *sc = dev->si_drv1;
   if (sc == NULL)
   {
      return ENXIO;
   }
   int revents = 0;
   sx_xlock(&(sc->sxlock));
   mtx_lock(&(sc->isrlock));
   if ((sc->state == CX88_VIDEO_CAPTURING) && (events & (POLLIN | POLLRDNORM)))
   {
      if (sc->channels_avail)
      {
         /* Notify the application that data is available */
         revents = events & (POLLIN | POLLRDNORM);
      }
      else
      {
         selrecord(td, &(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   sx_xunlock(&(sc->sxlock));
   return revents;
}

int cx88_video_filter(cx88_video_softc *sc)
{
	if (cx88_read_reg(sc->common, sc->stream->reg_intr_mstat) != 0)
	{
		return FILTER_SCHEDULE_THREAD;
	}
	return FILTER_STRAY;
}

void cx88_video_intr(cx88_video_softc *sc)
{   
   mtx_lock(&(sc->isrlock));
   /* Ensure we actually have an interrupt */
   uint32_t intrs = cx88_read_reg(sc->common, sc->stream->reg_intr_stat);
   if (intrs & sc->intr_mask)
   {
      if (sc->state != CX88_VIDEO_CAPTURING)
      {
         mtx_unlock(&(sc->isrlock));
         return;  
      }
      /* Check for fatal errors */
      if (intrs & sc->stream->bus_err_mask)
      {
         sc->isr_status |= CX88_VIDEO_BUS_ERROR;
      }
      for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
      {
         if (intrs & sc->stream->opcode_err_mask[i])
         {
            sc->isr_status |= CX88_VIDEO_DMA_PROGRAM;
         }
         if (intrs & sc->stream->sync_err_mask[i])
         {
            sc->isr_status |= CX88_VIDEO_SYNC_ERROR;
         }
         if (intrs & (sc->stream->risci1_mask[i] | sc->stream->risci2_mask[i]))
         {
            /* The counter indicates the most recently completed buffer */
            uint32_t counter = cx88_read_reg(sc->common, sc->stream->reg_gp_count[i]);
            uint32_t bufindex = counter;
            if (sc->format.in_even_field.enable && sc->format.in_odd_field.enable)
            {
               bufindex /= 2;
            }
            if (bufindex < sc->channels[i].num_buffers)
            {
               if (intrs & sc->stream->risci1_mask[i])
               {
                  sc->channel_configs[i].odd_frame_available = bufindex + 1;
                  sc->channels_avail |= CX88_VIDEO_FIELD_ODD(i);              
               }
               if (intrs & sc->stream->risci2_mask[i])
               {
                  if (!(sc->format.in_odd_field.enable) || ((counter % 2) == 1))
                  {
                     sc->channel_configs[i].even_frame_available = bufindex + 1;
                  }
                  else if (bufindex == 0)
                  {
                     sc->channel_configs[i].even_frame_available = sc->channels[i].num_buffers;
                  }
                  else
                  {
                     sc->channel_configs[i].even_frame_available = bufindex;
                  }
                  sc->channels_avail |= CX88_VIDEO_FIELD_EVEN(i);
               }
            }
         }
      }
      if (sc->isr_status & CX88_VIDEO_CAP_FATAL)
      {
         /* We should stop the transfer as soon as possible if we've encountered a fatal error */
         cx88_write_reg(sc->common, sc->stream->reg_dma_control, 0);
      }
      /* Acknowledge the interrupt */
      cx88_write_reg(sc->common, sc->stream->reg_intr_stat, intrs);
      if ((sc->channels_avail & sc->channel_mask) ||
          (sc->isr_status & CX88_VIDEO_CAP_FATAL))
      {
         wakeup(&(sc->channels_avail));
      }
      selwakeup(&(sc->selectors));
   }
   mtx_unlock(&(sc->isrlock));
}

static int cx88_video_lineprog(
   cx88_buffer *buffer,
   bus_size_t offset,
   cx88_video_clip_pattern *pattern)
{
   int error = 0;  
   uint16_t count = 0;
   boolean_t skip = FALSE;
   uint8_t i = 0;
   if ((pattern != NULL) && (pattern->in_clip_type == CX88_CLIP_EXCLUDE))
   {
      skip = TRUE;
   }
   do
   {
      uint16_t point = buffer->channel->line_size;
      if ((pattern != NULL) && (i < pattern->in_num_points))
      {
         point = pattern->in_points[i++];
      }
      if (point <= count)
      {
         device_printf(buffer->channel->sc->dev, "Out-of-order point %d in clipping pattern", point);
         return EINVAL;
      }
      if (point > buffer->channel->line_size)
      {
         point = buffer->channel->line_size;
      }
      uint16_t segcount = point - count;
      if (skip)
      {
         error = cx88_risc_skip_dmabuf(buffer, segcount, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
      }
      else
      {
         error = cx88_risc_transfer_dmabuf (buffer, offset + count,
            segcount, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
      }
      count += segcount;
      skip = !skip;
   } while (!error && (count < buffer->channel->line_size));
   return error;
}

static int cx88_video_fieldprog(
   cx88_buffer *buffer,
   cx88_video_channel_config *config,
   uint16_t *line,
   bus_size_t *offset,
   bus_size_t max_offset,
   bus_size_t line_size,
   uint8_t interlace)
{
   int error = 0;
   bus_size_t current_offset = *offset;
   uint16_t current_line = *line;
   cx88_video_clip_pattern *current_pattern = config->clip_patterns;
   uint32_t pattern_size = config->clip_pattern_size;
   while (!error && (current_offset < max_offset))
   {
      cx88_video_clip_pattern *pattern = NULL;
      while (pattern_size > sizeof(*current_pattern))
      {
         size_t current_size = 
            sizeof(*current_pattern) + (current_pattern->in_num_points * sizeof(uint16_t));
         if ((current_size > pattern_size) || (current_pattern->in_start_line > current_line))
         {
            break;    
         }
         else if (current_pattern->in_end_line >= current_line)
         {
            pattern = current_pattern;
            break;
         }
         current_pattern = (cx88_video_clip_pattern*)((uint8_t*)current_pattern + current_size);
         pattern_size -= current_size;
      }
      error = cx88_video_lineprog(buffer, current_offset, pattern);
      current_offset += (interlace * line_size);
      current_line += interlace;
   }
   *offset = current_offset;
   *line = current_line;
   return error;
}

static int cx88_video_dmaprog(cx88_buffer *buffer, uint8_t bufindex)
{
   int error = 0;
   cx88_video_softc *sc = buffer->dma_program_context;
   size_t channel_index = buffer->channel - sc->channels;
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Generating DMA program\n"))
   if (channel_index >= CX88_VIDEO_CHANNELS)
   {
      return EINVAL;
   }
   cx88_video_channel_config *config = &(sc->channel_configs[channel_index]);
   bus_size_t line_size = buffer->channel->line_size + config->x_header + config->x_trailer;
   bus_size_t current_offset = config->x_header;
   uint16_t line_number = 0;
   uint8_t interlace = (((sc->format.in_field_options & CX88_VIDEO_FIELD_INTERLACE) &&
                         (channel_index != CX88_VIDEO_CHANNEL_VBI)) ? 2 : 1);
   cx88_count_control countctl = ((bufindex == 0) ? CX88_RESET_COUNT : CX88_INCREMENT_COUNT);
   uint16_t odd_start_line = 0, even_start_line = 0x200;
   if (sc->format.in_field_options & CX88_VIDEO_FIELD_EVEN_FIRST)
   {
      odd_start_line = 0x200;
      even_start_line = 0;
   }
   if (sc->format.in_odd_field.enable)
   {
      bus_size_t max_offset = buffer->data_buffer_size;
      if (sc->format.in_even_field.enable && (interlace == 1))
      {
         max_offset /= 2;
      }
      cx88_risc_sync(buffer, odd_start_line, CX88_RISC_NO_SYNC_INTR, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
      error = cx88_video_fieldprog(buffer, config, &line_number,
         &current_offset, max_offset, line_size, interlace);
   }
   if (!error && (sc->format.in_even_field.enable))
   {
      if (sc->format.in_odd_field.enable)
      {
         cx88_risc_sync(buffer, even_start_line, CX88_RISC_NO_SYNC_INTR, CX88_TRIGGER_RISCI1, countctl);
         countctl = CX88_INCREMENT_COUNT;
      }
      else
      {
         cx88_risc_sync(buffer, even_start_line, CX88_RISC_NO_SYNC_INTR, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);   
      }
      if (interlace == 2)
      {
         current_offset = line_size + config->x_header;
         line_number = 1;
      }
      error = cx88_video_fieldprog(buffer, config, &line_number,
         &current_offset, buffer->data_buffer_size, line_size, interlace);
   }
   if (!error)
   {
      if ((sc->num_frames != 0) && (bufindex == (sc->num_frames - 1)))
      {
         cx88_risc_write_const_to_reg(
            buffer,
            0,
            sc->stream->dma_control_mask[channel_index] & VIDEO_FIFO_ENABLE_MASK, 
            sc->stream->reg_dma_control,
            CX88_NO_TRIGGER,
            CX88_IGNORE_COUNT);
      }
      
      /* Jump to next buffer, adjust the counter to reflect the index of the completed
      * buffer, and trigger RISC interrupt 1 so that the buffer can be processed */
      error = cx88_risc_next_buffer(
         buffer,
         0,
         (sc->format.in_even_field.enable ? CX88_TRIGGER_RISCI2 : CX88_TRIGGER_RISCI1),
         countctl);
   }
   return error;
}

static void cx88_video_reconfigure(cx88_video_softc *sc)
{
   uint8_t i;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      cx88_free_sram (&(sc->channels[i]));
   }
   if (sc->state > CX88_VIDEO_FORMATTED)
   {
      sc->state = CX88_VIDEO_FORMATTED;
   }
}

int cx88_video_request_reconfig(cx88_video_softc *sc)
{   
   int error = 0;
   /* Since this may be called from a sibling driver in a different thread
    * than any thread that is doing I/O with this driver, use trylock
    * a) because blocking on the lock could deadlock and
    * b) because the sibling wouldn't want to wait an indeterminate amount
    *    of time for I/O on this driver to complete */
   if (sx_try_xlock(&(sc->sxlock)))
   {
      if (sc->in_use)
      {
         /* If someone has a handle open to our device,
          * don't disrupt their session.  We'll just free our resources
          * when the handle is closed that they can be re-allocated
          * the next time a handle is opened. */
         sc->reconfig_requested = TRUE;
         error = EINPROGRESS;
      }
      else
      {
         cx88_video_reconfigure(sc);  
      }
      sx_xunlock(&(sc->sxlock));
   }
   else
   {
      sc->reconfig_requested = TRUE;
      error = EINPROGRESS;
   }
   return error; 
}

/* Suspend DEVMETHOD */
int cx88_video_suspend(cx88_video_softc *sc)
{
   sx_slock(&(sc->sxlock));
   cx88_video_reset(sc, CX88_VIDEO_UNCONFIGURED);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_video_release_channels(sc);
   sx_xunlock(&(sc->sxlock));
   return 0;
}

static int cx88video_mod_handler(module_t mod, int what, void *arg)
{
   switch(what)
   {
      case MOD_LOAD:
         cx88_video_units = new_unrhdr(0, INT_MAX, NULL);
         if (cx88_video_units == NULL)
         {
            return ENOMEM;
         }
         return 0;
      case MOD_UNLOAD:
         if (cx88_video_units != NULL)
         {
            delete_unrhdr(cx88_video_units);
            cx88_video_units = NULL;
         }
         return 0;
      default:
         return 0;
   }
   return 0;
}

/* Module declarations */
static moduledata_t cx88video_mod_data =
{
   "cx88videocore",
   cx88video_mod_handler,
   NULL
};

DECLARE_MODULE(cx88videocore, cx88video_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_DEPEND(cx88videocore, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);
MODULE_VERSION(cx88videocore, CX88_VIDEO_MODVER);
